from helper import eprint
import sys
import os
import time
import prover_runner as pr
import config
import argparse
from preprocessing import handle_pre_processing
import ltb_module
import helper as hp
import multiprocessing as mp


def main(problem_path, wc_timeout, schedule='default', schedule_mode='external',
         pre_processor_heuristic=None, pre_processor_timeout=200, no_cores=4, ltb=False,
         ltb_schedule_low_v4=None, ltb_schedule_low_v5=None, ltb_schedule_high=None, ltb_output_dir="."):

    # Check that problem/batchfile exists exists and exit if it doesn´t
    hp.check_file_existence(problem_path)

    # Get the current time as start time
    start_time = time.time()

    # Create the tmp dir
    hp.create_tmp_dir()

    # Add  grace
    time_left = wc_timeout + 5
    if config.VERBOSE:
        print("Added grace, new timelimit is: {0}".format(time_left))

    # Pre process problem if pre_processor is set (not avaliable for LTB)
    if pre_processor_heuristic and not ltb:
        problem_path = handle_pre_processing(
            problem_path,
            pre_processor_heuristic,
            pre_processor_timeout,
            time_left,
            ltb)

        # Prep did not solve the problem, update the time left and return the prep problem
        time_left = time_left - (time.time() - start_time)
        if config.VERBOSE:
            print("Time spent on prep: {0:.2f}".format(time.time() - start_time))
            print("Time left for prover: {0:.2f}".format(time_left))

    # Either run the LTB or standard setup
    if ltb:
        handle_ltb_run(
            problem_path,
            ltb_output_dir,
            schedule_mode,
            no_cores,
            ltb_schedule_low_v4,
            ltb_schedule_low_v5,
            ltb_schedule_high)
    else:
        handle_prover_run(problem_path, time_left, schedule, schedule_mode, no_cores)

    # Clean up and shut down
    # clean_tmp_folder()

    if config.VERBOSE:
        print("Finished")


def get_input_args():

    parser = argparse.ArgumentParser()

    # Positional arguments
    parser.add_argument("problem_path", help="Path to problem")
    parser.add_argument("wc_timeout", type=float, help="WC time limit for problem")

    # Optional Arguments
    parser.add_argument("--no_cores", type=int, help="No cores available for proving", default=8)
    parser.add_argument("--schedule", help="The schedule to run",
                        default="default", choices=list(config.SCHEDULES.keys()))
    parser.add_argument("--schedule_mode", help="How to run the heuristics on the prover",
                        default="external", choices=["external", "internal"])

    parser.add_argument("--pre_processor_heuristic", help="The heuristic to run as a pre-processor")
    parser.add_argument("--pre_processor_timeout", type=int,
                        help="The timeout for the pre-processor")

    # LTB arguments
    parser.add_argument(
        "--ltb",
        help="Set to run in LTB mode on BathcFile",
        action="store_true")
    parser.add_argument('--ltb_output_dir', help="Output directory for LTB proofs")
    parser.add_argument('--ltb_schedule_low_v4', help="The low time limit schedule for LTB (v4)",
                        choices=list(config.SCHEDULES.keys()))
    parser.add_argument('--ltb_schedule_low_v5', help="The low time limit schedule for LTB (v5)",
                        choices=list(config.SCHEDULES.keys()))
    parser.add_argument('--ltb_schedule_high', help="The high time limit schedule for LTB",
                        choices=list(config.SCHEDULES.keys()))

    # TODO integrate later
    #parser.add_argument("-v", "--verbosity", action="count", default=0, help="increase output verbosity")

    # Parse the arguments
    args = parser.parse_args()

    # Return the arguments
    return args


def check_illegal_args(args):

    try:
        if args.wc_timeout < 0:
            eprint("ERROR: wc_timeout cannot be negative")
            raise ValueError

        if args.pre_processor_timeout is not None and args.pre_processor_timeout < 0:
            eprint("ERROR: pre_processor_timeout cannot be negative")
            raise ValueError
        if args.pre_processor_timeout is not None and args.pre_processor_timeout > args.wc_timeout:
            eprint("ERROR: pre_processor_timeout cannot be larger than wc_timeout")
            raise ValueError
        if args.pre_processor_timeout is not None and args.pre_processor_heuristic is None:
            eprint("ERROR: Need to specify preprocessor heuristic (no default)")
            raise ValueError

        # LTB options
        if args.pre_processor_heuristic and args.ltb:
            eprint("ERROR: preprocessing currently not supported for LTB")
            raise ValueError
        # TODO quickfx
        #if args.ltb_schedule_low is not None and not args.ltb:
        #    eprint("ERROR: ltb_schedule_low only operational with ltb mode")
        #    raise ValueError
        if args.ltb_schedule_high is not None and not args.ltb:
            eprint("ERROR: ltb_schedule_high only operational with ltb mode")
            raise ValueError

    except ValueError:
        eprint("EXIT ON ERROR")
        sys.exit(1)


# Handler for standard prover runs (non-LTB)
def handle_prover_run(problem, time_left, schedule, schedule_mode, no_cores):

    # Start proving
    print("% SZS status Started for {0}".format(problem))

    # Run the prover setup
    success, output_file = run_prover_setup(problem, time_left, schedule, schedule_mode, no_cores)

    if success:
        # print("PROOF OUT SUPPRESSED")  # TODO
        hp.output_proof(output_file)
    else:
        print("% SZS status Unknown for {0}".format(problem))


def handle_ltb_run(batchfile_path, proof_output_dir, schedule_mode, no_cores,
                   ltb_schedule_low_v4, ltb_schedule_low_v5, ltb_schedule_high):

    # Check that both schedules are provided!!
    try:
        if ltb_schedule_low_v4 is None:
            eprint("ERROR: ltb_schedule_low_v4 is not provided!")
            raise ValueError()
        if ltb_schedule_low_v5 is None:
            eprint("ERROR: ltb_schedule_low_v5 is not provided!")
            raise ValueError()
        if ltb_schedule_high is None:
            eprint("ERROR: ltb_schedule_high is not provided!")
            raise ValueError()
    except ValueError:
        eprint("Exit on error")
        sys.exit(1)

    # Make the output directory
    if not os.path.exists(proof_output_dir):
        os.makedirs(proof_output_dir)

    # Get the start time
    start_time = time.time()

    # Get attributes of the batchfile
    batch_dir, time_limit, problems = ltb_module.process_batchfile(batchfile_path)

    # Create batch manager
    ltb_batch = ltb_module.create_ltb_batch(batch_dir, time_limit, problems)
    if config.VERBOSE:
        print("Loaded and created LTB batch: {0}".format(ltb_batch))

    # LTB LOW timelimit phase
    ltb_batch = run_ltb_low_timelimit_phase(ltb_batch, proof_output_dir, schedule_mode, ltb_schedule_low_v4, ltb_schedule_low_v5, config.LTB_PHASE_LOW_MAX_RUNTIME, no_cores)

    # LTB HIGH timelimit phase
    time_left = time_limit - (time.time() - start_time)
    ltb_batch = run_ltb_high_timelimit_phase(
        ltb_batch,
        proof_output_dir,
        schedule_mode,
        ltb_schedule_high,
        time_left,
        no_cores,
        version="+4")

    # If there for some reason is time left after running +4
    # We have a little attempt at +5
    # LTB HIGH timelimit phase
    time_left = time_limit - (time.time() - start_time)
    ltb_batch = run_ltb_high_timelimit_phase(
        ltb_batch,
        proof_output_dir,
        schedule_mode,
        ltb_schedule_high,
        time_left,
        no_cores,
        version="+5")

def run_ltb_high_timelimit_phase(ltb_batch, proof_output_dir, schedule_mode,
                                 ltb_schedule_high, timelimit, no_cores, version=None):

    # Get the start time
    start_time = time.time()

    # Get the list of unsolved problems
    unsolved_problems = ltb_batch.get_unsolved_problems()

    # Make a process pool
    pool = mp.Pool(processes=no_cores)

    # Create manager for accessing the number of unsolved problems while running proccesses
    # in parallell. The number of unsolved processes is used to calculate the time available.
    manager = mp.Manager()
    val = manager.Value('i', len(unsolved_problems))
    #val.value = len(problems)
    lock = manager.Lock()

    # Create function arguments for the problem attempt
    func_args = [(val, lock, schedule_mode, ltb_schedule_high, proof_output_dir, prob,
                  start_time, timelimit, no_cores, version) for prob in unsolved_problems]

    # Compute the processes
    res = pool.starmap(run_ltb_multi_schedule, func_args)

    # No more processes to the queue
    pool.close()
    # Wait for pool to finish processing before carrying on
    pool.join()

    if config.VERBOSE:
        print("All processes has finished")
        print("Final unsolved val: ", val.value)

        print("# # #")
        print("Process output results:")
        print(res)

    # Update the ltb_batch object
    for problem_name, success in res:
        if success:
            ltb_batch.set_problem_solved(problem_name)

    if config.VERBOSE:
        print("# " * 10)
        print(ltb_batch)

    return ltb_batch


def run_ltb_multi_schedule(no_unsolved, lock, schedule_mode, schedule, proof_output_dir, problem,
                           start_time, wc_timelimit, no_cores, version):

    # Get the problem version with the least number of clauses
    if version is None:
        problem_path, version_name = ltb_module.compute_optimal_problem_version_ltb(problem)
    else:
        problem_path, version_name = problem.get_path(version=version)

    print("% SZS status Started for {0}".format(version_name))

    # Compute the runtime for the attempt based on the time left
    # and the time available when sharing between the processes
    # and the other unsolved problems.
    # TODO this slightly overestimates the time left for the problem!
    time_left = wc_timelimit - (time.time() - start_time)
    time_available = min(time_left, (time_left * no_cores) / no_unsolved.value)
    runtime = max(time_available, config.LTB_PHASE_HIGH_MIN_RUNTIME)

    if config.VERBOSE:
        print("time_left: {0:.1f}".format(time_left))
        print("time_available: {0:.1f}".format(time_available))
        print("runtime: {0:.1f}".format(runtime))
        print("no_unsolved: {0}".format(no_unsolved.value))

    # Run the scheudle on the problem, but only limit to 1 core!
    success, proof_file = run_prover_setup(problem_path, runtime, schedule, schedule_mode, 1, ltb_mode=True)

    if success:
        print("% SZS status Theorem for {0}".format(version_name))
        # Print to file
        ltb_module.output_proof_to_file(proof_output_dir, problem.name, proof_file)
    else:
        print("% SZS status GaveUp for {0}".format(version_name))
    print("% SZS status Ended for {0}".format(version_name))

    # If solved, update the shared variable
    if success:
        with lock:
            no_unsolved.value -= 1

    # Return the problem name and the success status
    return problem.name, success


def run_ltb_low_timelimit_phase(ltb_batch, proof_output_dir, schedule_mode,
                                ltb_schedule_low_v4, ltb_schedule_low_v5, low_timelimit, no_cores):

    # Run low timelimit on +4 problems
    if config.VERBOSE:
        print("Starting LTB round: low-timelimit v:+4: timeout: {0}".format(low_timelimit))
    ltb_batch = run_ltb_problem_attempt(
        ltb_batch,
        proof_output_dir,
        schedule_mode,
        ltb_schedule_low_v4,
        low_timelimit,
        no_cores,
        version="+4")

    # Run low timelimit on +5 problems
    if config.VERBOSE:
        print("Starting LTB round: low-timelimit v:+5 timeout: {0}".format(low_timelimit))
    ltb_batch = run_ltb_problem_attempt(
        ltb_batch,
        proof_output_dir,
        schedule_mode,
        ltb_schedule_low_v5,
        low_timelimit,
        no_cores,
        version="+5")

    if config.VERBOSE:
        print("Low Time-limit phase ended: {0} problems solved".format(ltb_batch.get_no_solved_problems()))
        print()

    return ltb_batch


def run_ltb_problem_attempt(ltb_batch, proof_output_dir, schedule_mode,
                            schedule, attempt_timelimit, no_cores, version=None):

    for problem in list(ltb_batch.get_unsolved_problems()):

        # Get the problem_path from the problem
        problem_path, version_name = problem.get_path(version=version)
        print("% SZS status Started for {0}".format(version_name))

        success, output_file = run_prover_setup(
            problem_path, attempt_timelimit, schedule, schedule_mode, no_cores, ltb_mode=True)

        if success:
            print("% SZS status Theorem for {0}".format(version_name))
            ltb_batch = ltb_module.handle_solved_problem(ltb_batch, proof_output_dir, problem, output_file)
        else:
            print("% SZS status GaveUp for {0}".format(version_name))

        print("% SZS status Ended for {0}".format(version_name))

    return ltb_batch


def get_schedule_conf(schedule):

    # Get the schedule preprocessor QUICKFIX as this should be handled in the args handler
    try:
        schedule_conf = config.SCHEDULES[schedule]
    except KeyError:
        eprint("The schedule \"{0}\" is not implemented.".format(sys.argv[4]))
        sys.exit(1)

    return schedule_conf


def get_scheule_internal_prep_heuristic(schedule):

    # TODO move to the valid argparse checker
    # Get the schedule preprocessor QUICKFIX as this should be handled in the args handler
    try:
        internal_prep_heuristic = config.INTERNAL_PREP_HEURISTICS[schedule]
    except KeyError:
        eprint("The prep heuristic for \"{0}\" is not implemented.".format(schedule))
        sys.exit(1)

    return internal_prep_heuristic


def get_schedule_conf_internal(schedule):

    # Get the prep
    internal_prep_heuristic = get_scheule_internal_prep_heuristic(schedule)

    # Get the schedule list
    schedule_conf = get_schedule_conf(schedule)

    # The internal conf cosnists of: [prep_heur, schedule_heurs]
    schedule_conf = [internal_prep_heuristic, schedule_conf]

    return schedule_conf


def run_prover_setup(problem, time_left, schedule, schedule_mode, no_cores, ltb_mode=False):

    # Extract the schedule
    if schedule_mode == "external":
        schedule_conf = get_schedule_conf(schedule)
    elif schedule_mode == "internal":
        schedule_conf = get_schedule_conf_internal(schedule)
    else:
        eprint("ERROR: No schedule configurator determined for mode: {0}".format(schedule_mode))
        sys.exit(1)

    # Run iprover setup (internal or external scheduling)
    if schedule_mode == "internal":
        # TODO update the caller here! (#cores)
        success, proof_file = pr.run_schedule(problem, schedule_mode, schedule_conf, time_left, 1, ltb_mode)

    elif schedule_mode == "external":
        # Run the schedule externally
        success, proof_file = pr.run_schedule(
            problem, schedule_mode, schedule_conf, time_left, no_cores, ltb_mode)

    else:
        eprint("ERROR: schedule_mode {0} not implemented".format(schedule_mode))
        sys.exit(1)

    return success, proof_file


if __name__ == "__main__":

    # Get the arguments
    args = get_input_args()

    # Check for illegall arg combinations: This can probably be done smarter
    # in the argparses (later)
    check_illegal_args(args)

    # Run the program
    main(**vars(args))
