import subprocess
import config
from timeit import default_timer as time_now
from time import sleep
from helper import eprint
import os
import re
import signal
import time
import sys


# Object for storing prover processes
class ProverProc(object):

    def __init__(self, proc, output_file, start_time, max_runtime, heur_id="id none"):
        self.proc = proc
        self.output_file = output_file
        self.start_time = start_time
        self.max_runtime = max_runtime
        self.heur_id = heur_id
        #self.cmd_string = cmd_string

    def set_identifier(self, heur_id):
        self.heur_id = heur_id
        return self

    def __str__(self):
        return str(self.heur_id)


# Function for returning whether a terminated prover run was successfull or not
def check_prover_success(szs_status, ltb_mode):

    # LTB mode check (uses axiom selector)
    if ltb_mode:
        if re.match('(Theorem|Unsatisfiable)', szs_status):
            return True
        else:
            return False
    # Standard zsz ontology success check
    else:
        if re.match('(Theorem|Unsatisfiable|Satisfiable|CounterSatisfiable)', szs_status):
            return True
        else:
            return False


# Function for checking which processes in a process dictionary that is still running
# and terminated that process group (parent and children)
def kill_prover_process(prover_proc):

    # If process is still active
    if prover_proc.proc.poll() is None:
        try:
            # Get process group id and kill the group
            pgrp = os.getpgid(prover_proc.proc.pid)
            os.killpg(pgrp, signal.SIGKILL)

        except ProcessLookupError:
            # Process terminated before we could terminate it. The cpu time and proc_dict update
            # therefore does not occur. The terminated process should be handled by the the
            # terminated process handler function
            pass


# Get the running time and szs status of the output of a terminated prover run.
# The expected prover output is a bytestring.
def get_prover_output_status(prover_out):

    try:
        # Get the SZS status
        szs_status = re.search('% SZS status.*', prover_out).group(0)
        szs_status = szs_status.split()[3]

    except AttributeError:
        szs_status = "Unknown"

    return szs_status


# Kills running processes and updates the CPU times for cleaning up
def kill_all_prover_processes(proc_dict):

    # Kill each process
    for k, prover_proc in proc_dict.items():
        kill_prover_process(prover_proc)


def handle_finished_prover_run(prover_proc):
    # Handle terminated prover by inspecting the output and aquire the szs status

    # Make sure process is all finished
    try:
        _, errs = prover_proc.proc.communicate(timeout=10)  # timeout as the clausifier can hang
    except subprocess.TimeoutExpired:
        prover_proc.kill()
        _, errs = prover_proc.communicate()

    # Report any errors
    if errs != b"" and errs != b"\n\n":
        eprint(errs)

    # Read the proof output
    with open(prover_proc.output_file, 'r') as f:
        prover_out = f.read()

    # Parse the proof output
    szs_status = get_prover_output_status(prover_out)

    # Return run status
    return szs_status


# This function looks for terminated processes, and if they exist handles them.
# If a successfull prover run is found, the function terminates all other processes.
def checkTerminatedProcesses(proc_dict, provers_running, ltb_mode):

    # Loop through all the processes currently running
    proc_list = list(proc_dict.items())
    for dict_index, prover_proc in proc_list:

        # If process has finished
        if prover_proc.proc.poll() is not None:

            # Delete process from the running process overview
            del proc_dict[dict_index]

            # Decrement the number of processes running
            provers_running -= 1

            # Parse output form prover run
            szs_status = handle_finished_prover_run(prover_proc)

            if config.VERBOSE:
                print("Finished: {0} with status {1}".format(prover_proc, szs_status))

            # Get if prover was success and if so sat or unsat
            if check_prover_success(szs_status, ltb_mode):

                # Return prover has terminated with success
                return provers_running, True, proc_dict, prover_proc.output_file
            # TODO add else: remove the tmp file

        # Process has not finished but it should have timed out according to the timeout
        # we gave it. We include a 3 second grace here
        elif prover_proc.start_time + prover_proc.max_runtime < time.time() + 3.0:

            # Possibly "Overkill"
            kill_prover_process(prover_proc)
            prover_proc.proc.kill()
            _ = prover_proc.proc.communicate()  # Empty buffer

            # Remove process from the dictionary
            del proc_dict[dict_index]

            # Report
            if config.VERBOSE:
                print("Forcefully Terminated Process: {0}".format(prover_proc))

            # Decrement the number of processes running
            provers_running -= 1

    # No terminations with success
    return provers_running, False, proc_dict, ""


def check_if_last_call(no_cores, total_attempts, this_attempt_no):
    # The last processes to start on a core is the last call.
    # This is used to determine whether a process should run for its
    # assigned time, or for the time left, according to what best maximises
    # the time available.
    return total_attempts - this_attempt_no <= no_cores


def run_schedule(problem, schedule_mode, heuristic_spec, time_limit, no_cores, ltb_mode=False):

    # Set the start time
    start_time = time.time()

    # The current heuristic index
    heuristic_index = 0

    # The number of provers/processes currently running
    provers_running = 0

    # Dictionary for storing process information
    proc_dict = {}

    # Represents whether a prover has found a proof
    success = False

    # Start running prover processes for each core available.
    # We continue to start new prover processes while:
    #   - There are more heuristics left
    #   - The timeout for the problem is not met
    #   - A process has not terminated with success
    while heuristic_index < len(heuristic_spec) and time_now() - \
            start_time < time_limit and not success:

        # Compute the time left
        time_left = time_limit - (time.time() - start_time)

        # Compute whether the next process is a last call, -> should run until assigned completion
        last_call = check_if_last_call(no_cores, len(heuristic_spec), heuristic_index)

        # Build the cmd string according to the schedule
        if schedule_mode == "external":
            cmd_string, max_run_time, heur_id = build_prover_cmd_external(
                problem, time_left, heuristic_spec[heuristic_index], last_call=last_call)
        elif schedule_mode == "internal":
            cmd_string, max_run_time, heur_id = build_prover_cmd_internal(
                problem, time_left, heuristic_spec, last_call=last_call)
        else:
            eprint("ERROR: Mode {0} not defined for build_prover_cmd*".format(schedule_mode))
            sys.exit(1)

        # Start the prover process
        prover_proc = start_prover_process(cmd_string, max_run_time)
        # Set the id for recognition later
        prover_proc.set_identifier(heur_id)

        # Store the prover process in a process dict
        proc_dict[str(heuristic_index)] = prover_proc

        # Increment to the next heuristic
        heuristic_index += 1

        # Increment the number of provers running
        provers_running += 1

        # If we are utilising all the cores, have no successfull proofs and have not met the timelimit
        # we sleep until the above is not satisfied (most likely a prover terminates)
        while (provers_running >= no_cores or heuristic_index == len(heuristic_spec)
               ) and not success and time_now() - start_time < time_limit and provers_running > 0:

            # Set program to sleep
            sleep(config.PROC_SLEEP_TIME)

            # Check if some processes have terminated
            provers_running, success, proc_dict, proof_file = checkTerminatedProcesses(
                proc_dict, provers_running, ltb_mode)

    # Finished running prover and heuristics, time to kill processes
    kill_all_prover_processes(proc_dict)

    # Return success
    return success, proof_file


def start_prover_process(cmd_string, max_runtime):

    # Direct stdout to a temporary file
    output_file = "tmp/{0}".format(time.time())
    cmd_string += '  > ' + str(output_file)

    if config.VERBOSE:
        print("Starting prover {0} for time {1:.2f}".format(config.PROVER, max_runtime))
        print("Running cmd: ", cmd_string)

    # Start the prover process
    proc = subprocess.Popen(cmd_string, stderr=subprocess.PIPE, shell=True, preexec_fn=os.setsid)

    # Store the information in a prover process object
    prover_proc = ProverProc(proc, output_file, time.time(), max_runtime)

    return prover_proc


def get_heuristic_string(heur_file, timeout, opt_filter=[]):

    # TODO check for clausification times

    # Get the heuristic from the config file if none is given
    # Open prover configuration and get heuristic

    # Read heuristic and make each option a list entry
    with open(heur_file, "r") as fp:
        op = fp.read().splitlines()

    #op = op.split("\n")
    # Split the option from the argument
    #op = [o.split(None, maxsplit=1) for o in op]
    #op = []


    # Add quatations to the list options and the clausifier options
    for n in range(len(op)):

        # Quotation values
        # First check nested queue
        if "[[" in op[n] and "]]" in op[n]:
            op[n] = op[n].replace("[[", "\"[[")
            op[n] = op[n].replace("]]", "]]\"")
        # Check normal lists
        elif "[" in op[n] and "]" in op[n]:
            op[n] = op[n].replace("[", "\"[")
            op[n] = op[n].replace("]", "]\"")
        elif "--clausifier_options" in op[n]:
            # Quote the clausifier options
            op[n] = op[n][:21] + "\"" + op[n][21:] + "\""
            # Change the timeout for the clausifier
            if "-t" in op[n]:
                op[n] = re.sub("-t [0-9]+([,.][0-9]+)?", "-t {0}".format(timeout), op[n])

        # As of 28-06/2020 we always make sure this options is set to/contains unflattening
        if "--sup_input_triv" in op[n]:
            #set_sup_input_triv = True
            # Cheeky replacement trick to ensure unflatting is always contained
            if op[n][-3] != "[":
                # Need to add semicolon if there are more arguments
                op[n] = op[n].replace("[", "[Unflattening;")
            else:
                op[n] = op[n].replace("[", "[Unflattening")


    # Always filter out the timeout
    opt_filter += ["--time_out_real"]

    # Remove entries which contains the filter index
    for opt in opt_filter:
        for n in range(len(op)):
            if opt in op[n]:
                del op[n]
                break  # Found the options, so exit

    # Concatenate all options as one string
    op = " ".join(op)

    return op


def build_prover_cmd_external(problem, time_left, heuristic_spec, last_call=False):

    # Compute the running time for this process
    heur_timelimit = heuristic_spec[1]
    if last_call or heur_timelimit is None:
        timeout = time_left
    else:
        timeout = min(heur_timelimit, time_left)

    # Get the heuristic in cmd string format
    heur_path = heuristic_spec[0]
    heur_string = get_heuristic_string(heur_path, timeout)

    # Set prover, heuristic and timeout
    cmd_string = "{0} {1} --time_out_real {2:.2f} {3}".format(config.PROVER, heur_string, timeout, problem)

    # Build a string identifier
    heur_id = "{0};{1:.2f}".format(heur_path, timeout)

    return cmd_string, timeout, heur_id


def build_prover_cmd_internal(problem, time_left, heuristic_spec, last_call=False):

    # Extract the prep heuristic
    prep_heuristic = heuristic_spec[0]

    # Get the internal schedule
    schedule_spec = heuristic_spec[1]

    # Staart with a runtime of 0 and incrementally build up
    timeout = 0

    # Build the internal schedule string
    internal_schedule = "["
    for n, (heur_name, heur_time) in enumerate(schedule_spec):

        # If this is the last heuristic or the time is None, we run with -1
        if n + 1 >= len(schedule_spec) and (last_call or heur_time is None):
            # This is the last heuristic and should run until timeout
            internal_schedule += "[{0};-1];".format(heur_name)

            # Make sure the timeout is larger than the time left such that it is truncated
            timeout += 2 * (time_left + 1)
        else:
            # Because of support of None for the external schedules, we have to  convert
            # the None into the time_left. This will be truncated in the overall timeout
            if heur_time is None:
                heur_time = time_left

            # Set the internal heuristic
            internal_schedule += "[{0};{1:.0f}];".format(heur_name, heur_time)

            # Increment the timeout with this heuristic
            timeout += heur_time

    # Remove the last semi-colon and append closing bracket
    if len(internal_schedule) > 2:
        internal_schedule = internal_schedule[:-1]
    internal_schedule += "]"

    # We usually spend 10% of the prover time on preprocessing
    timeout = int(timeout * 1.10)

    # Truncate the timeout, in case it exceeds the time available
    timeout = min(timeout, time_left)

    # Get the prep heuristic and make sure the --schedule options is removed
    prep_heur_string = get_heuristic_string(prep_heuristic, timeout, opt_filter=["schedule"])

    # Build the execution string
    # Set prover, prep heuristic, internal heuristic and timeout
    cmd_string = "{0} {1} --schedule \"{2}\" --time_out_real {3:.2f} {4}".format(
        config.PROVER, prep_heur_string, internal_schedule, timeout, problem)

    return cmd_string, timeout, internal_schedule
